@@ -1,7 +1,4 @@
-# ! /usr/bin/python
-# -*- coding: utf-8 -*-
-
-# Copyright (c) 2019, NVIDIA CORPORATION. All rights reserved.
+# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -15,6 +12,12 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import abc
+from abc import ABC
+from typing import Dict, Optional, Tuple
+
+from nemo.core.neural_types.comparison import NeuralTypeComparisonResult
+
 __all__ = [
     'ElementType',
     'VoidType',
@@ -26,6 +29,7 @@ __all__ = [
     'MFCCSpectrogramType',
     'LogitsType',
     'LabelsType',
+    'HypothesisType',
     'LossType',
     'RegressionValuesType',
     'CategoricalValuesType',
@@ -35,12 +39,26 @@ __all__ = [
     'EmbeddedTextType',
     'EncodedRepresentation',
     'MaskType',
+    'Target',
+    'ClassificationTarget',
+    'ImageFeatureValue',
+    'Index',
+    'ImageValue',
+    'NormalizedImageValue',
+    'StringLabel',
+    'StringType',
+    'TokenIndex',
+    'Length',
+    'IntType',
+    'FloatType',
+    'NormalDistributionSamplesType',
+    'NormalDistributionMeanType',
+    'NormalDistributionLogVarianceType',
+    'TokenDurationType',
+    'TokenLogDurationType',
+    'LogDeterminantType',
+    'SequenceToSequenceAlignmentType',
 ]
-import abc
-from abc import ABC, abstractmethod
-from typing import Dict, Optional, Tuple
-
-from nemo.core.neural_types.comparison import NeuralTypeComparisonResult
 
 
 class ElementType(ABC):
@@ -48,7 +66,7 @@ class ElementType(ABC):
     We are relying on Python for inheritance checking"""
 
     def __str__(self):
-        self.__doc__
+        return self.__doc__
 
     def __repr__(self):
         return self.__class__.__name__
@@ -92,6 +110,9 @@ class ElementType(ABC):
                 return NeuralTypeComparisonResult.SAME_TYPE_INCOMPATIBLE_PARAMS
             else:
                 for k1, v1 in self.type_parameters.items():
+                    if v1 is None or second.type_parameters[k1] is None:
+                        # Treat None as Void
+                        continue
                     if v1 != second.type_parameters[k1]:
                         return NeuralTypeComparisonResult.SAME_TYPE_INCOMPATIBLE_PARAMS
             # check that all fields match
@@ -135,6 +156,11 @@ class LabelsType(ElementType):
     a more concrete types such as RegressionValuesType, etc."""
 
 
+class HypothesisType(LabelsType):
+    """Element type to represent some decoded hypothesis, which may further be processed to obtain
+    a concrete label."""
+
+
 class LengthsType(ElementType):
     """Element type representing lengths of something"""
 
@@ -158,7 +184,7 @@ class AudioSignal(ElementType):
         freq is the same.
     """
 
-    def __init__(self, freq: int = 16000):
+    def __init__(self, freq: int = None):
         self._params = {}
         self._params['freq'] = freq
 
@@ -192,4 +218,103 @@ class CategoricalValuesType(PredictionsType):
 
 
 class MaskType(PredictionsType):
-    """Element type to represent boolean mask"""
+    """Element type to represent a boolean mask"""
+
+
+class Index(ElementType):
+    """Type representing an element being an index of the sample."""
+
+
+class Target(ElementType):
+    """
+        Type representing an element being a target value.
+    """
+
+
+class ClassificationTarget(Target):
+    """
+        Type representing an element being target value in the classification task, i.e. identifier of a desired class.
+    """
+
+
+class ImageValue(ElementType):
+    """
+        Type representing an element/value of a single image channel,
+        e.g. a single element (R) of RGB image.
+    """
+
+
+class NormalizedImageValue(ImageValue):
+    """
+        Type representing an element/value of a single image channel normalized to <0-1> range,
+        e.g. a single element (R) of normalized RGB image.
+    """
+
+
+class ImageFeatureValue(ImageValue):
+    """Type representing an element (single value) of a (image) feature maps."""
+
+
+class StringType(ElementType):
+    """Element type representing a single string"""
+
+
+class StringLabel(StringType):
+    """
+        Type representing an label being a string with class name (e.g. the "hamster" class in CIFAR100).
+    """
+
+
+class BoolType(ElementType):
+    """Element type representing a single integer"""
+
+
+class IntType(ElementType):
+    """Element type representing a single integer"""
+
+
+class FloatType(ElementType):
+    """Element type representing a single float"""
+
+
+class TokenIndex(IntType):
+    """Type representing an element being index of a token in some kind of a vocabulary."""
+
+
+class Length(IntType):
+    """Type representing an element storing a "length" (e.g. length of a list)."""
+
+
+class ProbabilityDistributionSamplesType(ElementType):
+    """Element to represent tensors that meant to be sampled from a valid probability distribution
+    """
+
+
+class NormalDistributionSamplesType(ProbabilityDistributionSamplesType):
+    """Element to represent tensors that meant to be sampled from a valid normal distribution
+    """
+
+
+class SequenceToSequenceAlignmentType(ElementType):
+    """Class to represent the alignment from seq-to-seq attention outputs. Generally a mapping from endcoder time steps
+    to decoder time steps."""
+
+
+class NormalDistributionMeanType(ElementType):
+    """Element to represent the mean of a normal distribution"""
+
+
+class NormalDistributionLogVarianceType(ElementType):
+    """Element to represent the log variance of a normal distribution"""
+
+
+class TokenDurationType(ElementType):
+    """Element for representing the duration of a token"""
+
+
+class TokenLogDurationType(ElementType):
+    """Element for representing the log-duration of a token"""
+
+
+class LogDeterminantType(ElementType):
+    """Element for representing log determinants usually used in flow models"""
